#!/usr/bin/env python
#
# Copyright (C) 2002 Gre7g Luterman <gre7g@wolfhome.com>
#
# This file is part of TMDA.
#
# TMDA is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.  A copy of this license should
# be included in the file COPYING.
#
# TMDA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with TMDA; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"Generic session handler."

import glob
import os
import pickle
import pwd
import random
import re
import string
import time
import CgiUtil
import Authenticate

Rands = random.Random()

# Constants
CGI_SESSION_PREFIX = "/tmp/TMDASess."

class Session:
  """Sessioning tool for use in CGI.

Resurrect an old session by passing a form with the session ID (SID) into the 
constructor.  An empty object will be created if the session has expired.  A new
session will be created if the form does not specify a (valid) SID.

Instantiating a session will check Form["user"] and Form["password"] if they 
exist to establish Session["UID"].  To check the validity of a session, test 
Session.has_key("UID") before any action that might reveal sensative 
information.

Once a session is validated, the module will switch EUID to Session["UID"].

The session's SID is saved as member SID.

Session data is stored in the object by treating it as a dictionary.  For
example:
  A = Session(Form)
  A['key'] = value
  A.Save()
  print "http://some/url?SID=%s" % A.SID

The Save() member saves the session's current values on disk, but can only be 
called once a validated session has been established.  Writes are done as 
CGI_USER, as specified in Defaults."""
  
  def __init__(self, Form):
    "Reload an existing SID or create a new one."
    
    global Defaults

    # Existing, valid looking session?
    if Form.has_key("SID") and \
      re.compile("^[a-zA-Z0-9]{8}$").search(Form["SID"].value):
      self.SID = Form["SID"].value

      # Resurrect session
      try:
        F = open(CGI_SESSION_PREFIX + self.SID)
        self.Vars = pickle.load(F)
        F.close()
        os.environ["HOME"] = self.Vars["HOME"]
        os.environ["USER"] = self.Vars["User"]
        
        # Is there a TMDARC variable?
        if os.environ.has_key("TMDARC"):
          # Yes, replace it
          os.environ["TMDARC"] = os.environ["TMDARC"].replace("/~/",
            "/%s/" % self.Vars["User"])
        
        # Now that we know who we are, get our defaults
        from TMDA import Defaults
        
        # Make sure the session has not been hijacked
        if os.environ["REMOTE_ADDR"] == self.Vars["IP"]: return
      except IOError: # Failed to resurrect session, fall through to make new SID
        pass
      
    # New session
    SessionChars = string.ascii_letters + string.digits
    self.SID = ""
    for i in range(8):
      self.SID += SessionChars[Rands.randrange(len(SessionChars))]
    self.Vars = {}

    # Logging in?
    if not Form.has_key("user"): return

    # Is there a TMDARC variable?
    if os.environ.has_key("TMDARC"):
      # Yes, replace it
      os.environ["TMDARC"] = os.environ["TMDARC"].replace("/~/",
        "/%s/" % Form["user"].value)

    # Validate the new session
    if Authenticate.CheckPassword(Form) > 0:
      self.Vars["User"]  = Form["user"].value
      PasswordRecord     = pwd.getpwnam(self.Vars["User"])
      self.Vars["UID"]   = PasswordRecord[2]
      self.Vars["GID"]   = PasswordRecord[3]
      self.Vars["HOME"]  = PasswordRecord[5]
      os.environ["USER"] = self.Vars["User"]
      os.environ["HOME"] = self.Vars["HOME"]
      self.Vars["IP"]    = os.environ["REMOTE_ADDR"]
      self.Vars["debug"] = Form["debug"].value

      # Now that we know who we are, get our defaults
      from TMDA import Defaults

      # Test CGI_ACTIVE
      if not Defaults.CGI_ACTIVE:
        CgiUtil.TermError("<tt>CGI_ACTIVE</tt> has not been set to 1.",
          "Incomplete configuration file.", "read configuration files",
          "%s<br>%s" %
          (CgiUtil.FileDetails("Global configuration", Defaults.GLOBAL_TMDARC),
          CgiUtil.FileDetails("Local configuration", Defaults.TMDARC)),
          """Add <tt>CGI_ACTIVE = 1</tt> to one of the configuration files or modify
file permissions to allow them to be read.""")

      # Get misc user info
      PasswordRecord  = pwd.getpwnam(Defaults.CGI_USER)
      self.Vars["MiscUID"] = PasswordRecord[2]
      self.Vars["MiscGID"] = PasswordRecord[3]

      self.Save() # Save session & set user

  def Save(self):
    """Save all session variables to disk and change user.  Possibly clean up 
old sessions."""
    # Change to CGI_USER
    try:
      os.seteuid(self.Vars["MiscUID"])
      os.setegid(self.Vars["MiscGID"])
    except OSError:
      pass
    F = open(CGI_SESSION_PREFIX + self.SID, "w")
    pickle.dump(self.Vars, F)
    F.close()

    # Clean up?
    if Rands.random() < Defaults.CGI_CLEANUP_ODDS:
      # Go through all sessions and check a-times
      Sessions = glob.glob(CGI_SESSION_PREFIX + "*")
      for Session in Sessions:
        try: # these commands could fail if another thread cleans simultaneously
          Stats = os.stat(Session)
          # Expired?
          if Stats[7] + Defaults.CGI_SESSION_EXP < time.time():
            os.unlink(Session)
        except OSError:
          pass

    # Change back to regular user
    try:
      os.seteuid(0)
      os.setegid(0)
      os.setegid(self.Vars["GID"])
      os.seteuid(self.Vars["UID"])
    except OSError:
      pass

  def __contains__(self, a): return a in self.Vars
  def countOf(self): return len(self.Vars)
  def __delitem__(self, a): del self.Vars[a]
  def __getitem__(self, a): return self.Vars[a]
  def __setitem__(self, a, b): self.Vars[a] = b
  def keys(self): return self.Vars.keys()
  def has_key(self, a): return self.Vars.has_key(a)
